# ##### BEGIN GPL LICENSE BLOCK #####
#
#  keI2M (Image To Mesh) - Import & convert (alpha) images to mesh
#  Copyright (C) 2023  Kjell Emanuelsson (ke-code.xyz)
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see https://www.gnu.org/licenses
#
# ##### END GPL LICENSE BLOCK #####

import bpy
import bmesh
import sys
import time
import os
import numpy as np
from bpy.types import PropertyGroup, Operator, Panel, AddonPreferences
from bpy_extras.io_utils import ImportHelper
from bpy.props import (
        BoolProperty,
        PointerProperty,
        FloatProperty,
        StringProperty,
        EnumProperty,
        IntProperty,
        FloatVectorProperty,
        CollectionProperty,
        )

bl_info = {
    "name": "kei2m Pro",
    "author": "Kjell Emanuelsson",
    "category": "Import-Export",
    "version": (2, 0, 0, 0),
    "blender": (2, 80, 0),
    "location": "Viewport / N-Panel / kei2m",
    "warning": "",
    "description": "Image(s) To Mesh Generator - Pro Version",
    "doc_url": "https://ke-code.xyz",
}

kei2m_version = 2.0


def load_slot(path):
    try:
        img = bpy.data.images.load(path, check_existing=True)
    except (OSError, IOError, Exception) as e:
        print("I2M Load Slot Error:\n", e)
        img = None
    return img


class KeI2M(Operator):
    bl_idname = "ke.i2m"
    bl_label = "kei2m"
    bl_description = "Image(s) To Mesh Generator"
    bl_options = {'REGISTER', 'UNDO'}

    opacity: IntProperty(min=1, max=100, default=25, name="Alpha Tolerance", subtype="PERCENTAGE", soft_min=25,
                         soft_max=25, description="100% = Will only allow completely opaque image pixels\n"
                                                  "Do not use as slider - Use keyboard input!")

    dilation: IntProperty(min=-99, max=99, default=0, name="Expand Border", soft_min=0, soft_max=0,
                          description="Expand Alpha Border (in pixels, roughly) - Zero to disable.\n"
                                      "(or Contract - with negative values)\n "
                                      "Tip: Tweak tolerance value 1st - only use Expand if necessary (slow!)")

    workres: EnumProperty(
        items=[("128", "128", "", "", 1),
               ("256", "256", "", "", 2),
               ("512", "512", "", "", 3),
               ("1024", "1024", "", "", 4),
               ("2048", "2048", "", "", 5),
               ("0", "Image", "", "", 6)],
        name="Work Resolution", default="256",
        description="The internal res. used for conversion. Use as low value as possible.\n"
                    "Warning: 1024+ can be VERY slow. (Alpha complexity also increase time)\n"
                    "Note: 'Image' will use the largest source axis as workres (Be careful)")

    obj_size: FloatProperty(min=0, default=1, name="Size", precision=3, soft_min=1, soft_max=1, subtype="DISTANCE",
                            description="Set custom size (longest axis) relative to the image, including alpha.\n"
                                        "Zero sets default (1m)")
    custom_pixel: FloatProperty(
        min=0, max=1, default=0, name="Custom Pixel Size", precision=5, soft_min=0, soft_max=0, subtype="DISTANCE",
        description="Set custom pixel size (Zero disables the override)\nObj Size = Pixel Size * Res")

    smooth_mode: EnumProperty(
        items=[("NONE", "None", "", "", 1),
               ("FAST", "Fast", "", "", 2),
               ("SMOOTH", "Smooth", "", "", 3)],
        name="Mesh Reduction", default="FAST",
        description="Which type of smoothing to use:\n"
                    "None   : Fastest & No smoothing. VERY high polycount: 1 pixel = 1 face!\n"
                    "Fast   : Fast & Simple smoothing (Medium poly count) \n"
                    "Smooth : Slower & Better smoothing (Low poly count)")

    material: EnumProperty(
        items=[("NONE", "None", "", "", 1),
               ("UVPROJ", "UV Project", "", "", 2),
               ("UVBAKE", "UV Applied", "", "", 3),
               ("VCOLOR", "VertexColor", "", "", 4)],
        name="Material", default="UVPROJ",
        description="None  : Image only used in mesh generation\n"
                    "UV-Prj: Material with UV-Projection Modifier & Empty\n"
                    "UV    : UV's are 'baked' (No modifiers/Empty)\n"
                    "VColor: No UVs. 1 face = 1 pixel (Disables smoothing)\n")

    reset: BoolProperty(default=False, options={"SKIP_SAVE", "HIDDEN"})
    batch: BoolProperty(default=False, options={"SKIP_SAVE", "HIDDEN"})

    rgba = []
    alpha_mask = []
    temp_img = None
    src_img = None
    obj = None
    width = 0
    height = 0
    ctx = None
    time = 0
    coll = None
    wm = None
    ref_res = None
    work_res = 256  # converted from workres
    vcolor = False  # set from material
    alpha_tolerance = 0.5  # converted from opacity
    overwrite = False
    sqcomp = [1, 1]
    # Prefs
    rgb_as_alpha = False
    user_rgb = (1, 0, 1)
    use_img_alpha = True

    @classmethod
    def poll(cls, context):
        return context.mode == "OBJECT"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        layout.prop(self, "opacity")
        layout.prop(self, "dilation")
        layout.prop(self, "workres", expand=True)
        row = layout.row(align=True)
        if self.custom_pixel != 0:
            row.enabled = False
        row.prop(self, "obj_size")
        layout.prop(self, "custom_pixel")
        row = layout.row(align=True)
        if self.material == "VCOLOR":
            row.enabled = False
        row.prop(self, "smooth_mode", expand=True)
        layout.prop(self, "material", expand=True)
        row = layout.row(align=False)
        split = row.split(factor=0.25)
        split.operator("wm.operator_defaults", icon="FILE_REFRESH", text="Reset")
        split.label(text=" ")
        layout.separator()

    def get_image_data(self):
        has_alpha = True if self.src_img.depth == 32 else False

        # Even if not using alpha for material, 'NONE' breaks the script (temp setting:)
        if has_alpha and self.src_img.alpha_mode != 'PREMUL':
            self.src_img.alpha_mode = 'PREMUL'

        # Resize temp work-img to workres
        self.temp_img = self.src_img.copy()
        size = self.temp_img.size[:]

        if self.work_res == 0:
            self.width, self.height = size[0], size[1]
            if size[0] >= size[1]:
                self.work_res = self.width
            else:
                self.work_res = self.height

        # Longest img axis used for workres
        if size[0] >= size[1]:
            self.width = self.work_res
            ratio = self.temp_img.size[1] / self.temp_img.size[0]
            self.height = int(self.work_res * ratio)
            self.ref_res = self.width
            self.sqcomp = [ratio, 1]
        else:
            ratio = self.temp_img.size[0] / self.temp_img.size[1]
            self.width = int(self.work_res * ratio)
            self.height = self.work_res
            self.ref_res = self.height
            self.sqcomp = [1, ratio]

        # Non-Square scale comp when using dilate (req. square), rescale in bmop later
        if self.dilation == 0:
            self.sqcomp = [1, 1]
        else:
            self.width = self.height = self.ref_res

        # Scale to workres & Store
        self.temp_img.scale(width=self.width, height=self.height)
        self.rgba = np.empty(self.width * self.height * 4, dtype=np.float32)
        self.temp_img.pixels.foreach_get(self.rgba)

    def make_alpha_mask(self):
        self.rgba.shape = (-1, 4)
        self.alpha_mask = np.where(self.rgba[:, 3] >= self.alpha_tolerance, 1, 0)
        if self.dilation != 0:
            self.dilate_alpha()

    def make_rgb_alpha_mask(self):
        self.rgba.shape = (-1, 4)
        rgbtol = self.alpha_tolerance * 0.5  # fits better
        # To-do: figure out faster np method?
        rmin = self.user_rgb[0] - rgbtol
        rmax = self.user_rgb[0] + rgbtol
        gmin = self.user_rgb[1] - rgbtol
        gmax = self.user_rgb[1] + rgbtol
        bmin = self.user_rgb[2] - rgbtol
        bmax = self.user_rgb[2] + rgbtol
        self.alpha_mask = []
        for rgba in self.rgba:
            if rmin < rgba[0] < rmax and gmin < rgba[1] < gmax and bmin < rgba[2] < bmax:
                self.alpha_mask.append(False)
            else:
                self.alpha_mask.append(True)
        self.alpha_mask = np.array(self.alpha_mask)
        if self.dilation != 0:
            self.dilate_alpha()

    def dilate_alpha(self):
        # Fast & 'Simple' Convolve Expand or Contract (this one only works with square dims?)
        v = abs(self.dilation)
        a = np.reshape(self.alpha_mask, (self.width, self.height))
        k = np.ones(v) / v
        a = np.apply_along_axis(lambda m: np.convolve(m, k, mode="same"), 1, a)
        a = np.apply_along_axis(lambda m: np.convolve(m, k, mode="same"), 0, a)
        if self.dilation > 0:
            a[a != 0] = 1
        self.alpha_mask = a.ravel()

    def create_obj(self, mesh_name):
        # 'Mask out' faces with no alpha to y-x grid indices (later x & z)
        mask_grid = np.reshape(self.alpha_mask, (self.height, self.width))
        # Note: invert (yx) needed for FRONT orientation (re-inv l8r) .T faster than list(zip(a[0), a[1]))?
        face_loops = np.asarray(np.where(mask_grid == 1)).T

        # Note: using co's as dict-keys is iffy but works w integers (scl resize w bmop l8r)
        offset = round(self.width * 0.5)
        faces = []
        verts = []
        verts_dict = {}

        def index_vert(co):
            idx = verts_dict.get(co, None)
            if idx is None:
                verts.append(co)
                idx = len(verts) - 1
                verts_dict[co] = idx
            return idx

        if self.vcolor:
            rgba = self.rgba[self.alpha_mask == 1]
            self.rgba = []
            for v, color in zip(face_loops, rgba):
                x = v[1] - offset
                y = v[0]
                xw = x + 1
                yh = y + 1
                v1 = index_vert((x, 0, y))
                v2 = index_vert((xw, 0, y))
                v3 = index_vert((xw, 0, yh))
                v4 = index_vert((x, 0, yh))
                faces.append((v1, v2, v3, v4))
                self.rgba.append(color)
        else:
            for y, x in face_loops:
                x -= offset
                xw = x + 1
                yh = y + 1
                v1 = index_vert((x, 0, y))
                v2 = index_vert((xw, 0, y))
                v3 = index_vert((xw, 0, yh))
                v4 = index_vert((x, 0, yh))
                faces.append((v1, v2, v3, v4))

        mesh = bpy.data.meshes.new(mesh_name)
        mesh.from_pydata(verts, [], faces)
        mesh.validate()
        self.obj = bpy.data.objects.new(mesh_name, mesh)

    def clean_mesh(self, o):
        bm = bmesh.new()
        bm.from_mesh(o.data)

        # Scaling down from 'integer-space' (faster w bmop!) & compensating if squareified
        scl = 1 / self.ref_res
        bmesh.ops.scale(bm, vec=(scl * self.sqcomp[1], 1, scl * self.sqcomp[0]), verts=bm.verts)

        if self.smooth_mode == "FAST":
            # Simple 'reverse 2XSAI' style smooth-reduction
            s_verts = [v for v in bm.verts if v.is_boundary and len(v.link_edges) <= 2]
            if s_verts:
                bmesh.ops.dissolve_verts(bm, verts=s_verts, use_face_split=True, use_boundary_tear=True)
                bmesh.ops.unsubdivide(bm, verts=bm.verts, iterations=10)

        elif self.smooth_mode == "SMOOTH":
            # Somehow this is the fastest method with decent smoothing(+reduction)?
            s_verts = []
            for f in bm.faces:
                es = []
                for e in f.edges:
                    if e.is_boundary:
                        es.extend(e.verts[:])
                visited = set()
                dup = {v for v in es if v in visited or (visited.add(v) or False)}
                s_verts.extend(dup)
            s_verts = list(set(s_verts))

            if s_verts:
                bmesh.ops.smooth_vert(bm, verts=s_verts, factor=1.0, use_axis_x=True, use_axis_y=False, use_axis_z=True)
                bmesh.ops.unsubdivide(bm, verts=bm.verts, iterations=32)
                bmesh.ops.dissolve_limit(bm, angle_limit=0.08727, verts=bm.verts, edges=bm.edges)

        bm.to_mesh(o.data)
        bm.free()

    def apply_vcolor(self, mesh):
        vc = mesh.vertex_colors.new()
        for f, c in zip(mesh.polygons, self.rgba):
            for loop in f.loop_indices:
                vc.data[loop].color = c

    def make_material(self, name, img):
        if not self.batch:
            m = bpy.data.materials.get(name)
            if m:
                bpy.data.materials.remove(m)
        m = bpy.data.materials.new(name=name)
        m.use_nodes = True
        shader = m.node_tree.nodes["Material Output"].inputs[0].links[0].from_node
        n_color = m.node_tree.nodes.new("ShaderNodeTexImage")
        m.node_tree.links.new(shader.inputs["Base Color"], n_color.outputs[0])
        n_color.image = img
        n_color.location = (-300, 300)
        if self.use_img_alpha:
            m.node_tree.links.new(shader.inputs["Alpha"], n_color.outputs[1])
        else:
            self.src_img.alpha_mode = 'NONE'
        return m

    def make_projector(self, obj, name):
        pname = name + "_PRJ"
        if self.overwrite:
            existing = bpy.data.objects.get(pname)
            if existing:
                bpy.data.objects.remove(existing)
        w = ((self.width / self.ref_res) * 0.5) * self.sqcomp[1]
        h = ((self.height / self.ref_res) * 0.5) * self.sqcomp[0]
        projector = bpy.data.objects.new(pname, None)
        self.coll.objects.link(projector)
        projector.hide_viewport = True
        projector.empty_display_type = 'IMAGE'
        projector.empty_display_size = self.obj_size * 2
        projector.rotation_euler[0] = 1.5707963
        projector.location[2] = h
        projector.scale = (w, h, 1)
        projector.parent = obj
        uv_project = obj.modifiers.new(name="I2M UV-Project", type="UV_PROJECT")
        uv_project.uv_layer = "UVMap"
        uv_project.projector_count = 1
        uv_project.projectors[0].object = projector

    def link_obj(self, obj):
        self.coll.objects.link(obj)
        self.ctx.view_layer.objects.active = obj
        obj.select_set(True)

    def progress_update(self, txt, done):
        if done:
            seconds = time.time() - self.time
            m = int(seconds // 60)
            s = round((seconds - (m * 60)), 3)
            if m == 0:
                msg = "%s Finished: %ss\n" % (txt, s)
            else:
                msg = "%s Finished: %sm %ss\n" % (txt, m, s)
        else:
            msg = txt
        sys.stdout.write(msg)
        sys.stdout.flush()

    def execute(self, context):
        self.ctx = context

        if self.reset:
            self.opacity = 25
            self.dilation = 0
            self.workres = "256"
            self.obj_size = 1
            self.custom_pixel = 0
            self.smooth_mode = "FAST"
            self.material = "UVPROJ"
            return {"FINISHED"}

        k = context.scene.kei2m
        kap = context.preferences.addons[__package__].preferences

        self.src_img = bpy.data.images[k.IMG]

        if not self.src_img:
            self.report({"INFO"}, "Image not found")
            return {"CANCELLED"}

        self.rgb_as_alpha = kap.use_rgb
        if self.rgb_as_alpha:
            self.user_rgb = kap.user_rgb

        self.overwrite = kap.overwrite

        # Batch setup ( since it refuses to reuse last used settings ran as op?..)
        if self.batch:
            self.opacity = k.opacity
            self.dilation = k.dilation
            self.workres = k.workres
            self.obj_size = k.obj_size
            self.custom_pixel = k.custom_pixel
            self.smooth_mode = k.smooth_mode
            self.material = k.material

        self.work_res = int(self.workres)
        self.alpha_tolerance = float(self.opacity / 100)

        if not self.obj_size:
            self.obj_size = 1

        # Get current collection / scene fallback
        cvl = context.view_layer.active_layer_collection.name
        collections = [c.name for c in bpy.data.collections]
        if cvl in collections:
            self.coll = bpy.data.collections[cvl]
        else:
            self.coll = context.scene.collection

        if self.material == "VCOLOR":
            self.vcolor = True
            self.smooth_mode = "NONE"
        else:
            self.vcolor = False

        # Mouse progress meter: Fake progress feedback (just to show things are happening)
        if not self.batch:
            self.wm = context.window_manager
            self.wm.progress_begin(0, 99)
            self.wm.progress_update(98)

        # Console window std print out
        self.time = time.time()
        self.progress_update("I2M: Processing...", done=False)

        # for each , batch images support TBD
        self.get_image_data()

        if self.rgb_as_alpha:
            self.make_rgb_alpha_mask()
        else:
            self.make_alpha_mask()

        # Create mesh object
        new_obj_name = "I2M_" + self.src_img.name.split('.')[0]

        if self.overwrite:
            existing_mesh = bpy.data.meshes.get(new_obj_name)
            # existing_obj = bpy.data.objects.get(new_obj_name)
            if existing_mesh:
                bpy.data.meshes.remove(existing_mesh)
            # if existing_obj:
            #     bpy.data.objects.remove(existing_obj)

        self.create_obj(new_obj_name)

        # Clean up mesh
        self.clean_mesh(self.obj)

        # Vertex color or UV projection material
        if self.vcolor:
            self.apply_vcolor(self.obj.data)
        else:
            # Add UV-Map
            self.obj.data.uv_layers.new(name="UVMap")

            # Add Material
            material_name = new_obj_name + "_MAT"
            mat = self.make_material(material_name, self.src_img)
            self.obj.data.materials.append(mat)

            # Add UV-Projector
            self.make_projector(self.obj, new_obj_name)

        self.link_obj(self.obj)

        if self.material == "UVBAKE":
            mod = self.obj.modifiers[0]
            override = {"object": self.obj}
            bpy.ops.object.modifier_apply(override, modifier=mod.name)
            bpy.data.objects.remove(self.obj.children[0], do_unlink=True)

        elif self.material == "NONE":
            if self.obj.children:
                bpy.data.objects.remove(self.obj.children[0], do_unlink=True)
            if self.obj.data.materials:
                for m in self.obj.data.materials:
                    bpy.data.materials.remove(m, do_unlink=True)
            if self.obj.modifiers:
                override = {"object": self.obj}
                for m in self.obj.modifiers:
                    bpy.ops.object.modifier_apply(override, modifier=m.name)

        if self.custom_pixel:
            self.obj_size = self.custom_pixel * self.work_res

        if self.obj_size != 1:
            self.obj.scale.x = self.obj_size
            self.obj.scale.z = self.obj_size

        self.progress_update("", done=True)

        if not self.batch:
            self.wm.progress_end()

        # Cleanup
        if self.temp_img is not None:
            bpy.data.images.remove(self.temp_img)

        # QoL
        if context.space_data.shading.type in {"SOLID", "RENDERED", "MATERIAL"}:
            if self.vcolor:
                context.space_data.shading.color_type = 'VERTEX'
            else:
                context.space_data.shading.color_type = 'TEXTURE'

        # Store settings for batch use (or reset if batch)
        if self.batch:
            self.batch = False
        else:
            # saving for batch operation
            k.opacity = self.opacity
            k.dilation = self.dilation
            k.workres = self.workres
            k.obj_size = self.obj_size
            k.custom_pixel = self.custom_pixel
            k.smooth_mode = self.smooth_mode
            k.material = self.material

        return {'FINISHED'}


class KeI2Mreload(Operator):
    bl_idname = "ke.i2m_reload"
    bl_label = "I2M Reload Image"

    def execute(self, context):
        k = context.scene.kei2m
        img = bpy.data.images.get(k.IMG)
        if img:
            img.reload()
            context.area.tag_redraw()
        else:
            self.report({"WARNING"}, "Failed. '%s' missing/empty/invalid" % k.IMG)
        return {"FINISHED"}


class KeI2Mfilebrowser(Operator, ImportHelper):
    bl_idname = "ke.i2m_filebrowser"
    bl_label = "Load Image"
    bl_description = "Opens a filebrowser to load an image"
    bl_options = {'REGISTER', 'UNDO'}

    filter_glob: StringProperty(
        default='*.png;*.tif;*.tiff;*.exr;*.hdr;*.tga;*.sgi;*.rgb;*.bw;*.jp2;*.j2c;*.cin;*.dpx;'
                '*.jpg;*.jpeg;*.bmp;*.gif;*.jxl;*.avif;*.webp;',
        options={'HIDDEN'})

    def execute(self, context):
        k = context.scene.kei2m
        kap = context.preferences.addons[__package__].preferences
        alpha_missing = ""

        # Trim extension (note: glob includes 'future' formats, just in case support is added)
        loaded = bpy.path.basename(self.filepath).split(".")
        loaded_name = loaded[0]

        # Filename check
        if len(loaded_name) > 63:
            print("I2M: Filename: '%s'" % loaded_name)
            print("  -> Filename is too long (>64) - will be shortened (internally)\n"
                  "  -> 'Reload Image(s)' operator will not work\n")

        img = load_slot(self.filepath)

        if img is not None:
            filename = img.name
            k.IMG = filename
            context.area.tag_redraw()
            if filename is not None:
                if img.depth != 32 and not kap.use_rgb:
                    alpha_missing = filename
                else:
                    print("I2M: Image Loaded: %s" % filename)
            else:
                self.report({"WARNING"}, "Could not load file (See console)")

        if alpha_missing:
            self.report({"WARNING"}, "Missing Alpha Channel: %s" % alpha_missing)
        else:
            if kap.autorun:
                if context.mode != "OBJECT":
                    bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.ke.i2m('INVOKE_DEFAULT', True)

        return {'FINISHED'}


class KeI2Mbatchbrowser(Operator, ImportHelper):
    bl_idname = "ke.i2m_batchbrowser"
    bl_label = "Batch Process"
    bl_description = "Select MULTIPLE IMAGES with the Batch filebrowser to process in sequence.\n" \
                     "Toggle 'Use Dir' option to use the entire folder/directory instead.\n" \
                     "Note: I2M will use previous settings = Run ONE image first to set."

    files: CollectionProperty(type=PropertyGroup)
    use_dir: BoolProperty(name="Use Dir", default=False, description="Process ALL images in current directory")

    def execute(self, context):
        if not self.filepath:
            return {"CANCELLED"}

        if not self.files and not self.use_dir:
            sys.stdout.write("I2M: Batch Process Aborted: No images loaded\n")
            self.report({"INFO"}, "Aborted: No images loaded")
            return {"CANCELLED"}

        # print(self.use_dir, " Folder name:", os.path.dirname(self.filepath))
        images = []
        img_count = 0
        k = context.scene.kei2m

        if self.use_dir:
            # globbed to avoid issues. Using all alpha + common formats for "rgb as alpha" (&future formats?)
            filter_glob = (
                '.png', '.tif', '.tiff', '.exr', '.hdr', '.tga', '.sgi', '.rgb', '.bw', '.jp2', '.j2c', '.cin',
                '.dpx', '.jpg', '.jpeg', '.bmp', '.gif', '.jxl', '.avif', '.webp')

            for file in os.listdir(self.filepath):
                if file.lower().endswith(filter_glob):
                    path = os.path.join(self.filepath, file)
                    img = load_slot(path)
                    if img is not None:
                        img_count += 1
                        images.append(img)
        else:
            directory = os.path.dirname(self.filepath)
            for file in self.files:
                path = os.path.join(directory, file.name)
                img = load_slot(path)
                if img is not None:
                    img_count += 1
                    images.append(img)

        sys.stdout.write("I2M: Batch Process Images Loaded: %s\n" % str(img_count))
        # Batch all loaded images
        for img in images:
            k.IMG = img.name
            bpy.ops.ke.i2m(batch=True)

        # Clear Slot
        k.IMG = ""

        return {'FINISHED'}


class VIEW3D_PT_i2m(Panel):
    bl_label = 'keI2M v%s Pro' % str(kei2m_version)
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'kei2m'

    def draw_header_preset(self, context):
        layout = self.layout
        layout.emboss = 'NONE'
        row = layout.row(align=False)
        row.prop(context.scene.kei2m, 'info', text="", icon="QUESTION", icon_only=True, emboss=False)
        row.separator(factor=0.5)

    def draw(self, context):
        k = context.scene.kei2m
        kap = context.preferences.addons[__package__].preferences

        img_loaded = bool(k.IMG)

        layout = self.layout
        col = layout.column(align=True)
        row = col.row(align=True)
        row.label(text="Load Image")
        row = col.row(align=True)
        row.operator("ke.i2m_filebrowser", text="", icon="FILEBROWSER")
        row.prop(k, "IMG", text="")
        row.operator("ke.i2m_reload", text="", icon="LOOP_BACK")

        col = layout.column(align=True)
        row = col.row(align=True)
        row.alignment = "LEFT"
        row.prop(k, "show_options",
                 icon="DOWNARROW_HLT" if k.show_options else "RIGHTARROW",
                 icon_only=False, text="Preferences", emboss=False
                 )
        if k.show_options:
            box = col.box().column(align=False)
            box.prop(kap, "autorun")
            box.prop(kap, "overwrite")
            box.prop(kap, "use_rgb", text="Use RGB as Alpha")
            if kap.use_rgb:
                box.prop(kap, "user_rgb", text="")

        box = layout.box()
        col = box.column(align=True)
        row = col.row(align=True)
        row.operator("ke.i2m_batchbrowser", text="Batch", icon="FILE_FOLDER")
        row.separator(factor=0.2)
        row.operator("ke.i2m", text="Reset I2M").reset = True

        col.separator(factor=0.25)
        row = col.row()
        row.scale_y = 1.8
        if not img_loaded:
            row.enabled = False
        row.operator("ke.i2m", text="Image To Mesh")


# Panels to update - tab-relocation
panels = (
        VIEW3D_PT_i2m,
        )


def update_panel(self, context):
    message = "I2M: Panel update failed"
    try:
        for panel in panels:
            if "bl_rna" in panel.__dict__:
                bpy.utils.unregister_class(panel)

        for panel in panels:
            panel.bl_category = context.preferences.addons[__package__].preferences.category
            bpy.utils.register_class(panel)

    except Exception as e:
        print("\n[{}]\n{}\n\nError:\n{}".format(__name__, message, e))
        pass


class KeI2Maddonprefs(AddonPreferences):
    bl_idname = __package__

    category: StringProperty(
            name="Tab Category",
            description="Choose a name (category) for tab placement",
            default="kei2m",
            update=update_panel
            )
    use_rgb: BoolProperty(name="Use RGB as Alpha", default=False,
                          description="Use the user-defined RGB color instead of an alpha channel")
    user_rgb: FloatVectorProperty(name="RGB Value", subtype="COLOR", size=3, default=(1.0, 1.0, 1.0))
    autorun: BoolProperty(name="AutoRun", default=False,
                          description="Automatically run I2M after loading image(s)")
    overwrite: BoolProperty(name="Overwrite", default=False,
                            description="Replace mesh from same image\nAvoids dupes in repeated use")

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.label(text="Tab Location (Category):")
        row.prop(self, "category", text="")
        col = layout.column(align=True)
        col.prop(self, "autorun")
        col.prop(self, "overwrite")
        row = layout.row(align=True)
        row.prop(self, "use_rgb")
        row.prop(self, "user_rgb", text="")


class KeI2Mprops(PropertyGroup):
    info: BoolProperty(
        name="kei2m General Info", default=False,
        description="1. Load Alpha Image (in Object Mode) with the folder icon under 'Load Image'\n"
                    "2. Click the 'Image To Mesh' button to convert.\n"
                    "3. Adjust settings to your liking in REDO PANEL.\n"
                    " - See Preferences for preset options\n")
    IMG: StringProperty(default="", name="I2M Image", description="")
    opacity: IntProperty(default=25)
    dilation: IntProperty(default=0)
    workres: StringProperty(default="256")
    obj_size: FloatProperty(default=1)
    custom_pixel: FloatProperty(default=0)
    smooth_mode: StringProperty(default="FAST")
    material: StringProperty(default="UVPROJ")
    show_options: BoolProperty(name="Options", default=False, description="Quick Access to I2M Prefs")


classes = (
    KeI2M,
    VIEW3D_PT_i2m,
    KeI2Maddonprefs,
    KeI2Mprops,
    KeI2Mfilebrowser,
    KeI2Mreload,
    KeI2Mbatchbrowser,
)


def register():
    for c in classes:
        bpy.utils.register_class(c)

    bpy.types.Scene.kei2m = PointerProperty(type=KeI2Mprops)

    # Force Panel Udpate, for custom tab location.
    try:
        if "bl_rna" in VIEW3D_PT_i2m.__dict__:
            bpy.utils.unregister_class(VIEW3D_PT_i2m)
        VIEW3D_PT_i2m.bl_category = bpy.context.preferences.addons[__package__].preferences.category
        bpy.utils.register_class(VIEW3D_PT_i2m)
    except Exception as e:
        print("I2M: Register panel update failed:\n", e)


def unregister():
    for c in reversed(classes):
        bpy.utils.unregister_class(c)

    try:
        del bpy.types.Scene.kei2m

    except Exception as e:
        print('unregister fail:\n', e)
        pass
